@@ -30,17 +30,22 @@
 #include <libintl.h>
 #include <unistd.h>
 #include <netdb.h>
+#include <net/if.h>
+#include <ifaddrs.h>
 #include <gtk/gtk.h>
 #include <glade/glade.h>
 #include <gconf/gconf-client.h>
-#include <libgnome/libgnome.h>
-#include <libgnomeui/libgnomeui.h>
 #include <dbus/dbus-glib.h>
+#include <glib/gi18n.h>
 
 #ifdef VINO_ENABLE_KEYRING
 #include <gnome-keyring.h>
 #endif
 
+#ifdef VINO_ENABLE_LIBUNIQUE
+#include <unique/unique.h>
+#endif
+
 #define VINO_PREFS_DIR                    "/desktop/gnome/remote_access"
 #define VINO_PREFS_ENABLED                VINO_PREFS_DIR "/enabled"
 #define VINO_PREFS_PROMPT_ENABLED         VINO_PREFS_DIR "/prompt_enabled"
@@ -49,13 +54,14 @@
 #define VINO_PREFS_VNC_PASSWORD           VINO_PREFS_DIR "/vnc_password"
 #define VINO_PREFS_MAILTO                 VINO_PREFS_DIR "/mailto"
 #define VINO_PREFS_ICON_VISIBILITY        VINO_PREFS_DIR "/icon_visibility"
-#define VINO_PREFS_LOCAL_ONLY             VINO_PREFS_DIR "/local_only"
+#define VINO_PREFS_NETWORK_INTERFACE      VINO_PREFS_DIR "/network_interface"
 #define VINO_PREFS_ENCRYPTION             VINO_PREFS_DIR "/require_encryption"
 #define VINO_PREFS_USE_ALTERNATIVE_PORT   VINO_PREFS_DIR "/use_alternative_port"
 #define VINO_PREFS_ALTERNATIVE_PORT       VINO_PREFS_DIR "/alternative_port"
 #define VINO_PREFS_LOCK_SCREEN            VINO_PREFS_DIR "/lock_screen_on_disconnect"
+#define VINO_PREFS_DISABLE_BACKGROUND     VINO_PREFS_DIR "/disable_background"
 
-#define N_LISTENERS 12
+#define N_LISTENERS                       13
 
 #define VINO_DBUS_BUS_NAME  "org.gnome.Vino"
 #define VINO_DBUS_INTERFACE "org.gnome.VinoScreen"
@@ -67,7 +73,7 @@ typedef struct {
 
   GtkWidget   *dialog;
   GtkWidget   *writability_warning;
-  GtkWidget   *url_labels_box;
+  GtkWidget   *send_email_button;
   GtkWidget   *url_label;
   GtkWidget   *allowed_toggle;
   GtkWidget   *prompt_enabled_toggle;
@@ -78,12 +84,19 @@ typedef struct {
   GtkWidget   *icon_always_radio;
   GtkWidget   *icon_client_radio;
   GtkWidget   *icon_never_radio;
-  GtkWidget   *local_only_toggle;
+  GtkWidget   *network_interface_combox;
+  GtkWidget   *network_interface_label;
   GtkWidget   *encryption_toggle;
   GtkWidget   *use_alternative_port_toggle;
   GtkWidget   *alternative_port_entry;
   GtkWidget   *lock_screen_toggle;
-  GtkTooltips *tips;
+  GtkWidget   *disable_background_toggle;
+#ifdef VINO_ENABLE_LIBUNIQUE
+  UniqueApp   *app;
+#endif
+
+  DBusGConnection *connection;
+  DBusGProxy      *proxy_name, *proxy_port;
 
   guint        listeners [N_LISTENERS];
   int          n_listeners;
@@ -164,79 +177,149 @@ vino_preferences_dialog_update_for_allowed (VinoPreferencesDialog *dialog,
 {
   gtk_widget_set_sensitive (dialog->prompt_enabled_toggle, allowed);
   gtk_widget_set_sensitive (dialog->view_only_toggle,      allowed);
-  gtk_widget_set_sensitive (dialog->url_labels_box,        allowed);
+  gtk_widget_set_sensitive (dialog->send_email_button,     allowed);
   gtk_widget_set_sensitive (dialog->password_toggle,       allowed);
   gtk_widget_set_sensitive (dialog->password_box,          allowed ? dialog->use_password : FALSE);
   gtk_widget_set_sensitive (dialog->icon_always_radio,     allowed);
   gtk_widget_set_sensitive (dialog->icon_client_radio,     allowed);
   gtk_widget_set_sensitive (dialog->icon_never_radio,      allowed);
-  gtk_widget_set_sensitive (dialog->local_only_toggle,     allowed);
+  gtk_widget_set_sensitive (dialog->network_interface_combox, allowed);
+  gtk_widget_set_sensitive (dialog->network_interface_label,  allowed);
   gtk_widget_set_sensitive (dialog->encryption_toggle,     allowed);
   gtk_widget_set_sensitive (dialog->use_alternative_port_toggle, allowed);
   gtk_widget_set_sensitive (dialog->alternative_port_entry,      allowed &&
 			    gconf_client_get_bool (dialog->client, VINO_PREFS_USE_ALTERNATIVE_PORT, NULL));
   gtk_widget_set_sensitive (dialog->lock_screen_toggle,          allowed);
+  gtk_widget_set_sensitive (dialog->disable_background_toggle,   allowed);
+}
+
+static GSList *
+vino_preferences_load_network_interfaces (void)
+{
+  struct ifaddrs *myaddrs, *ifa;
+  GSList *list = NULL;
+
+  /* Translators: 'All' means 'All network interfaces' */
+  list = g_slist_append (list, g_strdup (_("All")));
+
+  getifaddrs (&myaddrs);
+  for (ifa = myaddrs; ifa != NULL; ifa = ifa->ifa_next)
+    {
+      if (ifa->ifa_addr == NULL || ifa->ifa_name == NULL || (ifa->ifa_flags & IFF_UP) == 0) 
+        continue;
+
+      if (ifa->ifa_addr->sa_family == AF_INET || ifa->ifa_addr->sa_family == AF_INET6)
+        {
+          if (g_slist_find_custom (list, ifa->ifa_name, (GCompareFunc)g_strcasecmp) == NULL)
+            {
+              list  = g_slist_append (list, g_strdup (ifa->ifa_name));
+            }
+        }
+    }
+
+  freeifaddrs (myaddrs);
+  return list;
 }
 
 static void
-vino_preferences_dialog_local_only_toggled (GtkToggleButton       *toggle,
-					    VinoPreferencesDialog *dialog)
+vino_preferences_dialog_network_interface_update_combox (VinoPreferencesDialog *dialog,
+                                                         const gchar *selected_iface,
+                                                         gboolean first_time)
 {
-  gboolean local_only;
+  GSList *list;
+  gint    i, sel;
+  gchar  *iface_check;
 
-  local_only = gtk_toggle_button_get_active (toggle);
+  list = vino_preferences_load_network_interfaces ();
 
-  gconf_client_set_bool (dialog->client, VINO_PREFS_LOCAL_ONLY, local_only, NULL);
+  if (selected_iface)
+    iface_check = g_strdup (selected_iface);
+  else
+    iface_check = gconf_client_get_string (dialog->client, VINO_PREFS_NETWORK_INTERFACE, NULL);
+
+  sel = 0;
+  for(i = 0; list; list = list->next, i++)
+    {
+      gchar *iface = list->data;
+
+      if (!iface)
+        continue;
+
+      if (first_time)
+	gtk_combo_box_append_text (GTK_COMBO_BOX (dialog->network_interface_combox), iface);
+            
+      if (iface_check && !g_strcasecmp (iface, iface_check))
+	  sel = i;
+
+      g_free (iface);
+    }
+
+  gtk_combo_box_set_active (GTK_COMBO_BOX (dialog->network_interface_combox), sel);
+  g_slist_free (list);
+  g_free (iface_check);
 }
 
 static void
-vino_preferences_dialog_local_only_notify (GConfClient           *client,
-					   guint                  cnx_id,
-					   GConfEntry            *entry,
-					   VinoPreferencesDialog *dialog)
+vino_preferences_dialog_network_interface_notify (GConfClient           *client,
+                                                  guint                  cnx_id,
+                                                  GConfEntry            *entry,
+                                                  VinoPreferencesDialog *dialog)
 {
-  gboolean local_only;
+  gchar       *old_iface;
+  const gchar *new_iface;
 
-  if (!entry->value || entry->value->type != GCONF_VALUE_BOOL)
+  if (!entry->value || entry->value->type != GCONF_VALUE_STRING)
     return;
 
-  local_only = gconf_value_get_bool (entry->value) != FALSE;
+  new_iface = gconf_value_get_string (entry->value);
+  old_iface = gtk_combo_box_get_active_text (GTK_COMBO_BOX (dialog->network_interface_combox));
 
-  if (local_only != gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (dialog->local_only_toggle)))
-    {
-      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (dialog->local_only_toggle), local_only);
-    }
+  if (old_iface && new_iface && g_strcasecmp (old_iface, new_iface) != 0)
+    vino_preferences_dialog_network_interface_update_combox (dialog, new_iface, FALSE);
+
+  g_free (old_iface);
 }
 
-static gboolean
-vino_preferences_dialog_setup_local_only_toggle (VinoPreferencesDialog *dialog)
+static void
+vino_preferences_dialog_network_interface_changed (gpointer              unused,
+                                                   VinoPreferencesDialog *dialog)
 {
-  gboolean local_only;
+  gchar *iface;
 
-  dialog->local_only_toggle = glade_xml_get_widget (dialog->xml, "local_only_toggle");
-  g_assert (dialog->local_only_toggle != NULL);
+  iface = gtk_combo_box_get_active_text (GTK_COMBO_BOX (dialog->network_interface_combox));
 
-  local_only = gconf_client_get_bool (dialog->client, VINO_PREFS_LOCAL_ONLY, NULL);
+  if (g_strcasecmp (iface, _("All")))
+    gconf_client_set_string (dialog->client, VINO_PREFS_NETWORK_INTERFACE, iface, NULL);
+  else
+    gconf_client_set_string (dialog->client, VINO_PREFS_NETWORK_INTERFACE, "", NULL);
 
-  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (dialog->local_only_toggle), local_only);
+  g_free (iface);
+}
 
-  g_signal_connect (dialog->local_only_toggle, "toggled",
-		    G_CALLBACK (vino_preferences_dialog_local_only_toggled), dialog);
+static void
+vino_preferences_dialog_setup_network_interface_combox (VinoPreferencesDialog *dialog)
+{
+  dialog->network_interface_combox = glade_xml_get_widget (dialog->xml, "network_interface_combox");
+  dialog->network_interface_label = glade_xml_get_widget (dialog->xml, "network_interface_label");
+  g_assert (dialog->network_interface_combox != NULL);
+  vino_preferences_dialog_network_interface_update_combox (dialog, NULL, TRUE);
 
-  if (!gconf_client_key_is_writable (dialog->client, VINO_PREFS_LOCAL_ONLY, NULL))
+  g_signal_connect (dialog->network_interface_combox, "changed",
+                    G_CALLBACK (vino_preferences_dialog_network_interface_changed), dialog);
+
+  if (!gconf_client_key_is_writable (dialog->client, VINO_PREFS_NETWORK_INTERFACE, NULL))
     {
-      gtk_widget_set_sensitive (dialog->local_only_toggle, FALSE);
+      gtk_widget_set_sensitive (dialog->network_interface_combox, FALSE);
+      gtk_widget_set_sensitive (dialog->network_interface_label, FALSE);
       gtk_widget_show (dialog->writability_warning);
     }
 
-  dialog->listeners [dialog->n_listeners] = 
-    gconf_client_notify_add (dialog->client,
-			     VINO_PREFS_LOCAL_ONLY,
-			     (GConfClientNotifyFunc) vino_preferences_dialog_local_only_notify,
-			     dialog, NULL, NULL);
+  dialog->listeners [dialog->n_listeners] =
+    gconf_client_notify_add (dialog->client, 
+                VINO_PREFS_NETWORK_INTERFACE,
+                (GConfClientNotifyFunc)vino_preferences_dialog_network_interface_notify,
+                dialog, NULL, NULL);
   dialog->n_listeners++;
-
-  return local_only;
 }
 
 static void
@@ -921,7 +1004,7 @@ vino_preferences_dialog_password_changed (GtkEntry              *entry,
     {
       gconf_client_unset (dialog->client, VINO_PREFS_VNC_PASSWORD, NULL);
     }
-  else if (strlen (password) > 1)
+  else
     {
       char *password_b64;
 
@@ -1002,6 +1085,67 @@ vino_preferences_dialog_setup_password_entry (VinoPreferencesDialog *dialog)
     }
 }
 
+static void
+vino_preferences_dialog_disable_background_toggled (GtkToggleButton       *toggle,
+                                                    VinoPreferencesDialog *dialog)
+{
+  gboolean disable_background;
+
+  disable_background = gtk_toggle_button_get_active (toggle);
+
+  gconf_client_set_bool (dialog->client, VINO_PREFS_DISABLE_BACKGROUND, disable_background, NULL);
+}
+
+static void
+vino_preferences_dialog_disable_background_notify (GConfClient           *client,
+                                                   guint                  cnx_id,
+                                                   GConfEntry            *entry,
+                                                   VinoPreferencesDialog *dialog)
+{
+  gboolean disable_background;
+
+  if (!entry->value || entry->value->type != GCONF_VALUE_BOOL)
+    return;
+
+  disable_background = gconf_value_get_bool (entry->value) != FALSE;
+
+  if (disable_background != gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (dialog->disable_background_toggle)))
+    {
+      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (dialog->disable_background_toggle), disable_background);
+    }
+}
+
+static gboolean
+vino_preferences_dialog_setup_disable_background_toggle (VinoPreferencesDialog *dialog)
+{
+  gboolean disable_background;
+
+  dialog->disable_background_toggle = glade_xml_get_widget (dialog->xml, "disable_background_toggle");
+  g_assert (dialog->disable_background_toggle != NULL);
+
+  disable_background = gconf_client_get_bool (dialog->client, VINO_PREFS_DISABLE_BACKGROUND, NULL);
+
+  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (dialog->disable_background_toggle), disable_background);
+
+  g_signal_connect (dialog->disable_background_toggle, "toggled",
+		    G_CALLBACK (vino_preferences_dialog_disable_background_toggled), dialog);
+
+  if (!gconf_client_key_is_writable (dialog->client, VINO_PREFS_DISABLE_BACKGROUND, NULL))
+    {
+      gtk_widget_set_sensitive (dialog->disable_background_toggle, FALSE);
+      gtk_widget_show (dialog->writability_warning);
+    }
+
+  dialog->listeners [dialog->n_listeners] = 
+    gconf_client_notify_add (dialog->client,
+			     VINO_PREFS_DISABLE_BACKGROUND,
+			     (GConfClientNotifyFunc) vino_preferences_dialog_disable_background_notify,
+			     dialog, NULL, NULL);
+  dialog->n_listeners++;
+
+  return disable_background;
+}
+
 static void
 vino_preferences_server_updated (DBusGProxy *proxy,
                                  const char *name,
@@ -1013,146 +1157,74 @@ vino_preferences_server_updated (DBusGProxy *proxy,
     vino_preferences_dialog_update_url_label ( (VinoPreferencesDialog *) user_data);
 }
 
+static void
+vino_preferences_server_port_changed (DBusGProxy *proxy, gpointer user_data)
+{
+  vino_preferences_dialog_update_url_label ( (VinoPreferencesDialog *) user_data);
+}
+
 static void
 vino_preferences_start_listening (VinoPreferencesDialog *dialog)
 {
-  DBusGConnection *connection;
-  GError          *error;
-  DBusGProxy      *proxy;
-  
-  error = NULL;
-  connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
-  if (!connection)
-    {
-      g_printerr (_("Failed to open connection to bus: %s\n"),
-                  error->message);
-      g_error_free (error);
-      return;
-    }
+  gchar       *obj_path;
+  GdkScreen   *screen;
 
-  proxy = dbus_g_proxy_new_for_name (connection,
-                                     DBUS_SERVICE_DBUS,
-                                     DBUS_PATH_DBUS,
-                                     DBUS_INTERFACE_DBUS);
-
-  dbus_g_proxy_add_signal (proxy, "NameOwnerChanged",
+  dialog->proxy_name = dbus_g_proxy_new_for_name (dialog->connection,
+						  DBUS_SERVICE_DBUS,
+						  DBUS_PATH_DBUS,
+						  DBUS_INTERFACE_DBUS);
+  dbus_g_proxy_add_signal (dialog->proxy_name, "NameOwnerChanged",
                            G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INVALID);
-  dbus_g_proxy_connect_signal (proxy, "NameOwnerChanged",
+  dbus_g_proxy_connect_signal (dialog->proxy_name, "NameOwnerChanged",
                           G_CALLBACK (vino_preferences_server_updated), dialog, NULL);
+
+  screen = gtk_window_get_screen (GTK_WINDOW (dialog->dialog));
+  obj_path = g_strdup_printf ("/org/gnome/vino/screens/%d",
+                              gdk_screen_get_number (screen));
+  dialog->proxy_port = dbus_g_proxy_new_for_name (dialog->connection,
+						  VINO_DBUS_BUS_NAME,
+						  obj_path,
+						  VINO_DBUS_INTERFACE);
+  g_free (obj_path);
+  dbus_g_proxy_add_signal (dialog->proxy_port, "ServerPortChanged", G_TYPE_INVALID);
+  dbus_g_proxy_connect_signal (dialog->proxy_port, "ServerPortChanged",
+                          G_CALLBACK (vino_preferences_server_port_changed), dialog, NULL);
 }
 
 static int
 vino_preferences_get_server_port (VinoPreferencesDialog *dialog)
 {
-#define VINO_DEFAULT_PORT   5900
-#define VINO_MIN_PORT       5000
-#define VINO_MAX_PORT       6000
-
-  DBusGConnection *connection;
-  GError          *error;
-  DBusGProxy      *proxy;
-  int              port;
-  GdkScreen       *screen;
-  char            *obj_path;
-  
-  error = NULL;
-  connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
-  if (!connection)
-    {
-      g_printerr (_("Failed to open connection to bus: %s\n"),
-                  error->message);
-      g_error_free (error);
-      return 0;
-    }
-
-  screen = gtk_window_get_screen (GTK_WINDOW (dialog->dialog));
-
-  obj_path = g_strdup_printf ("/org/gnome/vino/screens/%d",
-                              gdk_screen_get_number (screen));
-
-  proxy = dbus_g_proxy_new_for_name (connection,
-                                     VINO_DBUS_BUS_NAME,
-                                     obj_path,
-                                     VINO_DBUS_INTERFACE);
-
-  g_free (obj_path);
-
-  if (!dbus_g_proxy_call (proxy, "GetServerPort", &error,
-                          G_TYPE_INVALID,
-                          G_TYPE_INT, &port,
-                          G_TYPE_INVALID))
-    {
-      g_object_unref (proxy);
-      g_error_free (error);
-      dbus_g_connection_unref (connection);
-      return 0;
-    }
-
-  g_object_unref (proxy);
-  dbus_g_connection_unref (connection);
-
-  if (port >= VINO_MIN_PORT && port <= VINO_MAX_PORT)
-    return port - VINO_DEFAULT_PORT;
-  else
-    return port;
-
-#undef VINO_DEFAULT_PORT
-#undef VINO_MIN_PORT
-#undef VINO_MAX_PORT
-}
-
+  DBusGProxy  *proxy;
+  int          port;
+  GdkScreen   *screen;
+  char        *obj_path;
 #ifdef VINO_ENABLE_HTTP_SERVER
-static int
-vino_preferences_get_http_server_port (VinoPreferencesDialog *dialog)
-{
-#define VINO_HTTP_DEFAULT_PORT   5800
-  DBusGConnection *connection;
-  GError          *error;
-  DBusGProxy      *proxy;
-  int              port;
-  GdkScreen       *screen;
-  char            *obj_path;
+  const char  *method = "GetHttpServerPort";
+#else
+  const char  *method = "GetServerPort";
+#endif
   
-  error = NULL;
-  connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
-  if (!connection)
-    {
-      g_printerr (_("Failed to open connection to bus: %s\n"),
-                  error->message);
-      g_error_free (error);
-      return 0;
-    }
-
   screen = gtk_window_get_screen (GTK_WINDOW (dialog->dialog));
-
   obj_path = g_strdup_printf ("/org/gnome/vino/screens/%d",
                               gdk_screen_get_number (screen));
 
-  proxy = dbus_g_proxy_new_for_name (connection,
+  proxy = dbus_g_proxy_new_for_name (dialog->connection,
                                      VINO_DBUS_BUS_NAME,
                                      obj_path,
                                      VINO_DBUS_INTERFACE);
-
   g_free (obj_path);
 
-  if (!dbus_g_proxy_call (proxy, "GetHttpServerPort", &error,
+  if (!dbus_g_proxy_call (proxy, method, NULL,
                           G_TYPE_INVALID,
                           G_TYPE_INT, &port,
                           G_TYPE_INVALID))
     {
-      g_object_unref (proxy);
-      g_error_free (error);
-      dbus_g_connection_unref (connection);
-      return VINO_HTTP_DEFAULT_PORT;
+      port = 0;
     }
 
   g_object_unref (proxy);
-  dbus_g_connection_unref (connection);
-
   return port;
-#undef VINO_HTTP_DEFAULT_PORT
 }
-#endif /* VINO_ENABLE_HTTP_SERVER */
 
 static char *
 vino_preferences_get_local_hostname (void)
@@ -1187,18 +1259,21 @@ vino_preferences_dialog_get_server_command (VinoPreferencesDialog *dialog)
   char *server_url;
   int   server_port;
 
-  local_host  = vino_preferences_get_local_hostname ();
+  server_port = vino_preferences_get_server_port (dialog);
+
+  gtk_widget_set_sensitive (dialog->send_email_button, server_port);
+
+  if (server_port == 0)
+    return g_strdup (_("The service is not running"));
+
+  local_host = vino_preferences_get_local_hostname ();
   if (!local_host)
-    {
-      local_host = g_strdup_printf ("localhost");
-    }
+    local_host = g_strdup_printf ("localhost");
 
 #ifdef VINO_ENABLE_HTTP_SERVER
-  server_port = vino_preferences_get_http_server_port (dialog);
   server_url = g_strdup_printf ("http://%s:%d", local_host, server_port);
 #else
-  server_port = vino_preferences_get_server_port (dialog);
-  server_url = g_strdup_printf ("vncviewer %s:%d", local_host, server_port);
+  server_url = g_strdup_printf ("vinagre %s::%d", local_host, server_port);
 #endif
 
   g_free (local_host);
@@ -1207,35 +1282,36 @@ vino_preferences_dialog_get_server_command (VinoPreferencesDialog *dialog)
 }
 
 static char *
-vino_preferences_dialog_construct_mailto (VinoPreferencesDialog *dialog,
-					  const char            *url)
+vino_preferences_dialog_construct_mailto (VinoPreferencesDialog *dialog)
 {
   GString *mailto;
+  char *command;
+
+  command = vino_preferences_dialog_get_server_command (dialog);
 
   mailto = g_string_new ("mailto:");
   if (dialog->mailto)
     mailto = g_string_append (mailto, dialog->mailto);
 
   mailto = g_string_append_c (mailto, '?');
-  g_string_append_printf (mailto, "Body=%s", url);
+  g_string_append_printf (mailto, "Body=%s", command);
 
+  g_free (command);
   return g_string_free (mailto, FALSE);
 }
 
 static void
 vino_preferences_dialog_update_url_label (VinoPreferencesDialog *dialog)
 {
-  char *command;
-  char *mailto;
+  char *command, *label;
 
   command = vino_preferences_dialog_get_server_command (dialog);
-  mailto = vino_preferences_dialog_construct_mailto (dialog, command);
-
-  gtk_button_set_label (GTK_BUTTON (dialog->url_label), command);
-  gtk_link_button_set_uri (GTK_LINK_BUTTON (dialog->url_label), mailto);
+  label = g_strdup_printf ("<i>%s</i>", command);
+  
+  gtk_label_set_label (GTK_LABEL (dialog->url_label), label);
   
   g_free (command);
-  g_free (mailto);
+  g_free (label);
 }
 
 static void
@@ -1261,47 +1337,59 @@ vino_preferences_dialog_mailto_notify (GConfClient           *client,
     {
       g_free (dialog->mailto);
       dialog->mailto = g_strdup (mailto);
-      vino_preferences_dialog_update_url_label (dialog);
     }
 }
 
 static void
-vino_preferences_dialog_uri_hook (GtkLinkButton *button,
-                                  const gchar *link,
-                                  gpointer user_data)
+vino_preferences_email_button_clicked (GtkButton *button,
+				       VinoPreferencesDialog *dialog)
 {
   GError *error;
   GdkScreen *screen;
+  char *mailto, *command;
 
-  screen = gtk_widget_get_screen (GTK_WIDGET (button));
+  error   = NULL;
+  screen  = gtk_widget_get_screen (GTK_WIDGET (button));
+  mailto  = vino_preferences_dialog_construct_mailto (dialog);
+  command = vino_preferences_dialog_get_server_command (dialog);
 
-  error = NULL;
-  if (!gnome_url_show_on_screen (link, screen, &error))
+  if (!gtk_show_uri (screen, mailto, GDK_CURRENT_TIME, &error))
     {
-      /* FIXME better error handling!
-       *       What best to do? For the specific case
-       *       in this preferences dialog we want to be
-       *       able to pop up a dialog with the error
-       *       but also the vino URL as a selectable
-       *       label.
-       *
-       *       Maybe chain this up to the caller?
-       */
+      GtkWidget *message_dialog;
 
-      g_warning ("Failed to show URL '%s': %s\n",
-		 link, error->message);
+      message_dialog = gtk_message_dialog_new (GTK_WINDOW (dialog->dialog),
+					       GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
+					       GTK_MESSAGE_ERROR,
+					       GTK_BUTTONS_CLOSE,
+					       _("There was an error showing the URL \"%s\""),
+					       command);
+      gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (message_dialog),
+						"%s",
+						error->message);
+
+      gtk_window_set_resizable (GTK_WINDOW (message_dialog), FALSE);
+
+      g_signal_connect (message_dialog, "response",
+			G_CALLBACK (gtk_widget_destroy),
+			NULL);
+
+      gtk_widget_show (message_dialog);
       g_error_free (error);
     }
+
+  g_free (mailto);
+  g_free (command);
 }
 
 static void
 vino_preferences_dialog_setup_url_labels (VinoPreferencesDialog *dialog)
 {
-  char *command;
-  char *mailto;
+  char *command, *label;
+  GtkWidget *image;
 
-  dialog->url_labels_box = glade_xml_get_widget (dialog->xml, "url_labels_box");
-  g_assert (dialog->url_labels_box);
+  dialog->url_label = glade_xml_get_widget (dialog->xml, "url_label");
+  dialog->send_email_button = glade_xml_get_widget (dialog->xml, "send_email_button");
+  g_assert (dialog->url_label);
 
   dialog->listeners [dialog->n_listeners] = 
     gconf_client_notify_add (dialog->client,
@@ -1318,24 +1406,19 @@ vino_preferences_dialog_setup_url_labels (VinoPreferencesDialog *dialog)
     }
 
   command = vino_preferences_dialog_get_server_command (dialog);
-  mailto = vino_preferences_dialog_construct_mailto (dialog, command);
+  label = g_strdup_printf ("<i>%s</i>", command);
   
-  gtk_link_button_set_uri_hook (vino_preferences_dialog_uri_hook, NULL, NULL);
+  gtk_label_set_label (GTK_LABEL (dialog->url_label), label);
 
-  dialog->url_label = gtk_link_button_new_with_label (mailto, command);
-
-  dialog->tips = gtk_tooltips_new ();
-  gtk_tooltips_set_tip (dialog->tips, dialog->url_label,
-                        _("Send this command by email"), NULL);
-  g_object_ref_sink (dialog->tips);
+  image = gtk_image_new_from_icon_name ("gnome-stock-mail-fwd", GTK_ICON_SIZE_BUTTON);
+  gtk_button_set_image (GTK_BUTTON (dialog->send_email_button), image);
+  g_signal_connect (dialog->send_email_button,
+		    "clicked",
+		    G_CALLBACK (vino_preferences_email_button_clicked),
+		    dialog);
 
   g_free (command);
-  g_free (mailto);
-
-  gtk_box_pack_start (GTK_BOX (dialog->url_labels_box),
-		      dialog->url_label,
-		      FALSE, FALSE, 0);
-  gtk_widget_show (dialog->url_label);
+  g_free (label);
 }
 
 static void
@@ -1343,7 +1426,8 @@ vino_preferences_dialog_response (GtkWidget             *widget,
 				  int                    response,
 				  VinoPreferencesDialog *dialog)
 {
-  GError *error;
+  GError    *error;
+  GdkScreen *screen;
 
   if (response != GTK_RESPONSE_HELP)
     {
@@ -1351,8 +1435,10 @@ vino_preferences_dialog_response (GtkWidget             *widget,
       return;
     }
 
+  screen = gtk_widget_get_screen (widget);
   error = NULL;
-  gnome_help_display_desktop (NULL, "user-guide", "user-guide.xml", "goscustdesk-90", &error);
+
+  gtk_show_uri (screen, "ghelp:user-guide?goscustdesk-90", GDK_CURRENT_TIME, &error);
   if (error)
     {
       GtkWidget *message_dialog;
@@ -1391,6 +1477,7 @@ vino_preferences_dialog_init (VinoPreferencesDialog *dialog)
 
   const char *glade_file;
   gboolean    allowed;
+  GError     *error = NULL;
 
   dialog->expected_listeners = N_LISTENERS;
 
@@ -1418,6 +1505,15 @@ vino_preferences_dialog_init (VinoPreferencesDialog *dialog)
   dialog->client = gconf_client_get_default ();
   gconf_client_add_dir (dialog->client, VINO_PREFS_DIR, GCONF_CLIENT_PRELOAD_ONELEVEL, NULL);
 
+  dialog->connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
+  if (!dialog->connection)
+    {
+      g_printerr (_("Failed to open connection to bus: %s\n"),
+                  error->message);
+      g_error_free (error);
+      return FALSE;
+    }
+
   vino_preferences_dialog_setup_url_labels (dialog);
 
   dialog->writability_warning = glade_xml_get_widget (dialog->xml, "writability_warning");
@@ -1426,16 +1522,17 @@ vino_preferences_dialog_init (VinoPreferencesDialog *dialog)
 
   allowed = vino_preferences_dialog_setup_allowed_toggle (dialog);
 
-  vino_preferences_dialog_setup_prompt_enabled_toggle (dialog);
-  vino_preferences_dialog_setup_view_only_toggle      (dialog);
-  vino_preferences_dialog_setup_password_toggle       (dialog);
-  vino_preferences_dialog_setup_password_entry        (dialog);
-  vino_preferences_dialog_setup_icon_visibility       (dialog);
-  vino_preferences_dialog_setup_local_only_toggle     (dialog);
-  vino_preferences_dialog_setup_encryption_toggle     (dialog);
+  vino_preferences_dialog_setup_prompt_enabled_toggle       (dialog);
+  vino_preferences_dialog_setup_view_only_toggle            (dialog);
+  vino_preferences_dialog_setup_password_toggle             (dialog);
+  vino_preferences_dialog_setup_password_entry              (dialog);
+  vino_preferences_dialog_setup_icon_visibility             (dialog);
+  vino_preferences_dialog_setup_network_interface_combox    (dialog);
+  vino_preferences_dialog_setup_encryption_toggle           (dialog);
   vino_preferences_dialog_setup_alternative_port_entry      (dialog);
   vino_preferences_dialog_setup_use_alternative_port_toggle (dialog);
   vino_preferences_dialog_setup_lock_screen_toggle          (dialog);
+  vino_preferences_dialog_setup_disable_background_toggle   (dialog);
 
   g_assert (dialog->n_listeners == dialog->expected_listeners);
 
@@ -1443,6 +1540,10 @@ vino_preferences_dialog_init (VinoPreferencesDialog *dialog)
 
   gtk_widget_show (dialog->dialog);
 
+#ifdef VINO_ENABLE_LIBUNIQUE
+  unique_app_watch_window (dialog->app, GTK_WINDOW (dialog->dialog));
+#endif
+
   return TRUE;
 
 #undef VINO_GLADE_FILE  
@@ -1459,10 +1560,6 @@ vino_preferences_dialog_finalize (VinoPreferencesDialog *dialog)
     g_free (dialog->mailto);
   dialog->mailto = NULL;
 
-  if (dialog->tips)
-    g_object_unref (dialog->tips);
-  dialog->tips = NULL;
-
   if (dialog->client)
     {
       int i;
@@ -1484,6 +1581,48 @@ vino_preferences_dialog_finalize (VinoPreferencesDialog *dialog)
   if (dialog->xml)
     g_object_unref (dialog->xml);
   dialog->xml = NULL;
+
+#ifdef VINO_ENABLE_LIBUNIQUE
+  if (dialog->app)
+    g_object_unref (dialog->app);
+  dialog->app = NULL;
+#endif
+
+  if (dialog->proxy_name)
+    g_object_unref (dialog->proxy_name);
+  dialog->proxy_name = NULL;
+
+  if (dialog->proxy_port)
+    g_object_unref (dialog->proxy_port);
+  dialog->proxy_port = NULL;
+
+  if (dialog->connection)
+    dbus_g_connection_unref (dialog->connection);
+  dialog->connection = NULL;
+}
+
+static gboolean
+vino_preferences_is_running (VinoPreferencesDialog *dialog)
+{
+#ifdef VINO_ENABLE_LIBUNIQUE
+  dialog->app = unique_app_new ("org.gnome.Vino.Preferences", NULL);
+
+  if (unique_app_is_running (dialog->app))
+    {
+      UniqueResponse response;
+
+      response = unique_app_send_message (dialog->app, UNIQUE_ACTIVATE, NULL);
+        
+      g_object_unref (dialog->app);
+      dialog->app = NULL;
+       
+      return response == UNIQUE_RESPONSE_OK;
+    }
+  else
+    return FALSE;
+#else
+  return FALSE;
+#endif
 }
 
 int
@@ -1495,9 +1634,11 @@ main (int argc, char **argv)
   bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
   textdomain (GETTEXT_PACKAGE);
 
-  gnome_program_init (PACKAGE, VERSION, LIBGNOMEUI_MODULE,
-		      argc, argv, NULL);
+  gtk_init (&argc, &argv);
 
+  if (vino_preferences_is_running (&dialog))
+    return 0;
+  
   if (!vino_preferences_dialog_init (&dialog))
     {
       vino_preferences_dialog_finalize (&dialog);
